#!/usr/bin/env python
"""
Utility for simulating clay fabrics.  Not very sophisticated and has
no direct application to real sediments.

     Copyright (C) 2005  Kurt Schwehr

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

"""
__version__ = '$Revision: 1.5 $'.split()[1] # See man ident
__date__ = '$Date: 2005/11/02 21:41:55 $'.split()[1]
__author__ = 'Kurt Schwehr'

import math,sys
from pmag_verbosity import BOMBASTIC,VERBOSE,TRACE,TERSE,ALWAYS
from pmag_kds_version import VERSION

from units import deg2rad,rad2deg

import pylab # matplotlib
import struct

def writearray(filename,data,lineBetweenRow=False, breakInRow=None):
    """Ugly little data dumper

    lineBetweenRow - if true, puts a blank line between rows.  Good for gnuplot
    breakInRow - set to a number of points after which to make a line break
       e.g. 2: (x1,y1,x2,y2) -> x1 y1 newline x2 y2 
    """
    f = open(filename,'w')
    for row in data:
        for i in range(len(row)):
            f.write(str(row[i])+' ')
            if 0!=i and None!= breakInRow and i % breakInRow == 0:
                f.write('\n')
        f.write('\n')
        if lineBetweenRow: f.write('\n')
    return

class DevRandom:
    """
    Use /dev/random to get some random numbers.  Quality depends on OS.

    Probably good for random number seeds... that is what I am using
    it for.

    http://www.python.org/doc/current/lib/module-struct.html
    """
    def __init__(self):
        self.devrand = open ('/dev/random','r')
        return
    def getInt4(self):
        """return a 4 byte signed integer

        Approximately -2G to 2G
        """
        string = self.devrand.read(4)
        intval = struct.unpack('i',string)
        return intval[0]

def make_particles_uniform_dist_range_angle(density, width, maxDepth, minLength, maxLength,minAngle=5,maxAngle=20, verbosity=0):
    """
    Generate a random fabric of platy particles
    Uniform distribution...  angle is constrained to possible angles
    max angle in radians
    """
    if verbosity>=TRACE: print '# Making particles... ', width, maxDepth, minLength, maxLength
    import random;     r = DevRandom();    random.seed(r);    del(r)
    from math import pi, sin, cos
    particles=[]
    deltaAngle = maxAngle-minAngle
    assert(maxAngle>minAngle)
    minDepth = 0
    deltaDepth = density[1][0]-density[0][0] # FIX: should allow for depth steps to vary down core!
    centerFile = open ('center.dat','w')
    for d in density:
        if verbosity>=VERBOSE: print 'depth:',d[0],'   num of particles:',d[1]
        for particle in range(int(d[1])):
            xcenter = random.random() * width
            length = random.random() * (maxLength-minLength) + minLength
            l2 = length/2 # half on each side
            #angle = random.random() * 2* maxAngle  - maxAngle#pi
            depth = d[0] + random.random() * deltaDepth - deltaDepth/2
            fraction = 1-(depth-minDepth)/float(maxDepth-minDepth)
            angle = random.random() * (deltaAngle * fraction + minAngle)
            if random.random()>.5: angle = -angle # Flip half of them
            x1 = xcenter - l2 * cos(angle) 
            x2 = xcenter + l2 * cos(angle)
            centerFile.write(str(xcenter)+' '+str(-depth)+' '+str(rad2deg(angle))+'\n')
            y1 = depth - l2 * sin(angle)
            y2 = depth + l2 * sin(angle)
            particles.append((x1,y1,x2,y2))
    return particles


def make_particles_uniform_dist_max_angle(density, width, depth, minLength, maxLength,maxAngle=0, verbosity=0):
    """
    Generate a random fabric of platy particles
    Uniform distribution...  angle is constrained to possible angles
    max angle in radians
    """
    if verbosity>=TRACE: print '# Making particles... ', width, depth, minLength, maxLength
    import random;     r = DevRandom();    random.seed(r);    del(r)
    from math import pi, sin, cos
    particles=[]
    for d in density:
        if verbosity>=VERBOSE: print 'depth:',d[0],'   num of particles:',d[1]
        for particle in range(int(d[1])):
            xcenter = random.random() * width
            length = random.random() * (maxLength-minLength) + minLength
            l2 = length/2 # half on each side
            angle = random.random() * 2* maxAngle  - maxAngle#pi
            x1 = xcenter - l2 * cos(angle) 
            x2 = xcenter + l2 * cos(angle) 
            y1 = d[0] - l2 * sin(angle)
            y2 = d[0] + l2 * sin(angle)
            particles.append((x1,y1,x2,y2))
    return particles


def make_particles_uniform_dist(density, width, depth, minLength, maxLength,verbosity=0):
    """
    Generate a random fabric of platy particles

    Uniform distribution... any angle possible
    """
    if verbosity>=TRACE: print '# Making particles... ', width, depth, minLength, maxLength

    import random
    r = DevRandom()
    random.seed(r)
    del(r)

    from math import pi, sin, cos

    particles=[]

    for d in density:
        if verbosity>=VERBOSE: print 'depth:',d[0],'   num of particles:',d[1]
        for particle in range(int(d[1])):
            xcenter = random.random() * width
            length = random.random() * (maxLength-minLength) + minLength
            l2 = length/2 # half on each side
            angle = random.random() * pi
            #print ' ', xcenter, length, angle, rad2deg(angle)
            x1 = xcenter - l2 * cos(angle) 
            x2 = xcenter + l2 * cos(angle) 
            y1 = d[0] - l2 * sin(angle)
            y2 = d[0] + l2 * sin(angle)
            
            particles.append((x1,y1,x2,y2))
    return particles

def show_particles(density, width, depth, minLength, maxLength,paritcles,filename=None,verbosity=0):
    """
    matplotlib code

    FIX: It does not need min/maxLength!
    """
    pylab.xlabel('horizontal distance')
    pylab.ylabel('depth in core')

    pylab.plot((0,width,width,0,0),(0,0,depth,depth,0), hold=False)

    for grain in particles:
        x1,y1,x2,y2 = grain
        #print 'grain',grain,x1,y2
        col1=(x1,x2); col2=(y1,y2)
        pylab.plot(col1,col2,hold=True)

    pylab.axis([0,width,depth,0])
    if filename==None: pylab.show()
    else:
        if verbosity>=TRACE:
            print 'Saving figure to ',filename
        pylab.savefig(filename)
    return

def particles_gnuplot_dat(particles, width, minDepth, maxDepth, filename='fabricsim.dat',verbosity=0):
                          #, options=None):
    """
    Write out a gnuplot data file and some code to plot it

    filename - where to write the data file.  The gnuplot command file has .gp on the end
    """

    #print particles

    outDat = open (filename,'w')
    outDat.write('# gnuplot particle data file\n')
    outDat.write('#\n')
    outDat.write('# filename: '+filename+'\n')
    outDat.write('# cmd file: '+filename+'.gp\n')
    outDat.write('#\n')
    outDat.write('# number of particles: '+str(len(particles))+'\n')
    outDat.write('\n')

    #if options!=None:
        #print 'OPTIONS: ',type(options)
        #for key in options.keys():
        #    outDat.write('#'+key+': '+options['key']+'\n')
            
    # Write out all data points with a blank line between points
    for p in particles:
        outDat.write(str(p[0])+' '+str(p[1])+'\n'+str(p[2])+' '+str(p[3])+''+'\n\n')

    del(outDat)
    
    ########################################
    # Command file
    outCmd = open (filename+'.gp','w')
    outCmd.write('#!/usr/bin/env gnuplot\n')
    outCmd.write('#\n')

    outCmd.write('# filename: '+filename+'\n')
    outCmd.write('# cmd file: '+filename+'.gp\n')
    outCmd.write('#\n')
    outCmd.write('# number of particles: '+str(len(particles))+'\n')

    outCmd.write('set xlabel "width"\n')
    outCmd.write('set ylabel "depth(cm)"\n')
    outCmd.write('\n')
    outCmd.write('set xrange [0:'+str(width)+']\n') 
    outCmd.write('set yrange ['+str(maxDepth)+':'+str(minDepth)+']\n') # max min does the same as 'reverse'
    outCmd.write('set key off\n')
    outCmd.write('\n')
    outCmd.write('plot \''+filename+'\' with lines\n')
    outCmd.write('pause -1 "Hit return to continue"\n')
    outCmd.write('\n')
    outCmd.write('set terminal pdf\n')
    outCmd.write('set output \''+filename+'.pdf\'\n')
    outCmd.write('replot\n')
    outCmd.write('\n')
    outCmd.write('set terminal svg\n')
    outCmd.write('set output \''+filename+'.svg\'\n')
    outCmd.write('replot\n')
    outCmd.write('\n')
    outCmd.write('set terminal png\n')
    outCmd.write('set output \''+filename+'.png\'\n')
    outCmd.write('replot\n')
    outCmd.write('\n')

    import os
    os.chmod(filename+'.gp',0755)
    return

def const_density(width,depth,step, start):
    density = []
    import numarray
    for depth in numarray.arange(0,depth, step):
        density.append([depth, start])
    return density

def linear_density(width,maxDepth,step, start, end):
    "Create a density increasing linearly down core"
    density = []
    startDepth = 0
    import numarray
    for depth in numarray.arange(startDepth,maxDepth, step):
        fraction = (depth-startDepth)/float(maxDepth - startDepth)
        density.append([depth, start+(end-start)*fraction])
    return density

def printTodo():
    """
    Things that need to be done to this code:

    FIX: perterb the depth +/- 1/2 of depth step that the code is at
    FIX: Plot the density curve down the side of the plot
    FIX: Plot the compaction (angular) distribution down the other side of the plot
    FIX: Make the line color represent something like composition or magnetic properties
    FIX: Otherwise make all the particles the same color
    FIX: Allow multimodal distributions of size... e.g. clay size fraction, nano's, and sand size
    FIX: particle thickness
    """
    print printTodo.__doc__


if __name__ == "__main__":
    from optparse import OptionParser

    myparser = OptionParser(usage="%prog [options]",
                            version="%prog "+VERSION+' - '+__version__)

    myparser.add_option('--gnuplot',dest='gpFile',default=None,type='string',
                        help='Write out a gnuplot data file and a .gp command file')
    myparser.add_option('--save-density',dest='densityFile',default=None,type='string',
                        help='File to save the generated density to')

    myparser.add_option('--show',dest='show',default=False,action='store_true',
                        help='Show a matplotlib graph of the loop')
    myparser.add_option('--plot-file',dest='plotFile',default=None,type='string',
                        help='Save plot to a file.  Filename extension determines the filetype.  '+
                        'e.g. .svg, .png, .ps [NOTE: .svg good for Illustrator and .ps broken(?)]')

    myparser.add_option('--width',dest='width',default=100.,type='float',help='width of the plot[default: %default]')
    myparser.add_option('--depth',dest='depth',default=100.,type='float',help='[default: %default]')
    myparser.add_option('--minLength',dest='minLength',default=1.,type='float',help='[default: %default]')
    myparser.add_option('--maxLength',dest='maxLength',default=10.,type='float',help='[default: %default]')

    # FIX: how best to hangle all these choices?

    densityChoices = ('linear','exp','const')
    choicesStr = ''
    for i in range(len(densityChoices)): choicesStr+=densityChoices[i]+', '
    choicesStr=choicesStr[:-2]
    myparser.add_option('--density-type',dest='densityType',type='choice',choices=densityChoices,
                        default='linear',
                        help='What method to use to generate density [default: %default] [choices: '+choicesStr+']')

    angularDistChoices = ('uniform','limited_uniform','range_uniform')
    choicesStr = ''
    for i in range(len(angularDistChoices)): choicesStr+=angularDistChoices[i]+', '
    choicesStr=choicesStr[:-2]
    myparser.add_option('--angular-distribution',dest='angularDist',type='choice',choices=angularDistChoices,
                        default='uniform',
                        help='What method to use to generate angular dispersion '
                        +'[default: %default] [choices: '+choicesStr+']')


    myparser.add_option('--min-angle',dest='minAngle',default=90,type='float',
                        help='Maximum angle to use for a limited_uniform distribution (in degrees) [default: %default]')
    myparser.add_option('--max-angle',dest='maxAngle',default=90,type='float',
                        help='Maximum angle to use for a limited_uniform distribution (in degrees) [default: %default]')


    # Starting parameter
    myparser.add_option('--start',dest='start',default=1,type='float',help='[default: %default]')
    myparser.add_option('--end',dest='end',default=50,type='float',help='[default: %default]')

    myparser.add_option('--step',dest='step',default=1,type='float',help='How far between steps on the density graph [default: %default]')

    myparser.add_option('--todo',dest='todo',default=False, action='store_true',
                        help='Print todo list')

    import pmag_verbosity
    pmag_verbosity.addVerbosityOptions(myparser)

    (options,args) = myparser.parse_args()
    if options.verbosity > BOMBASTIC:  print options
    if options.todo:
        printTodo()
        sys.exit()

    density=None
    if options.densityType=='const':
        density=const_density(options.width,options.depth,options.step, options.start)
    elif options.densityType=='linear':
        density=linear_density(options.width,options.depth,options.step, options.start, options.end)
    elif options.densityType=='exp':
        assert(False)
    else:
        sys.exit('ERROR What kind of density type is that? ...'+str(options.densityType))

    if options.densityFile != None:
        writearray(options.densityFile,density)

#     steps = 50
#     #density = [ (x,x) for x in range(200) ]
#     density = [ [x/4.,x/4.] for x in range(int(options.depth)) ]

#     topOfOverpressure=50

#     from math import sin,pi
#     for i in range(steps):
#         frac = i/float(steps)
#         percent = 1 - .75*sin(pi*frac)

    
#     density[i+topOfOverpressure][1] = int(percent* density[i+topOfOverpressure][1])


#    particles = make_particles(density,options.width,options.depth,options.minLength,options.maxLength,
#                               verbosity=options.verbosity)

    particles = None
    if options.angularDist == 'uniform':
        particles = make_particles_uniform_dist(density,options.width,options.depth,
                                                options.minLength,options.maxLength,
                                                verbosity=options.verbosity)
    elif options.angularDist == 'limited_uniform':
        maxAngle = deg2rad(options.maxAngle)
        from math import pi
        assert(maxAngle<=pi)
        assert(maxAngle>=0)  # FIX: what really should be the maximum maxangle?
        particles = make_particles_uniform_dist_max_angle(density,options.width,options.depth,
                                                          options.minLength,options.maxLength,
                                                          maxAngle,
                                                          verbosity=options.verbosity)

    elif options.angularDist == 'range_uniform':
        maxAngle = deg2rad(options.maxAngle)
        minAngle = deg2rad(options.minAngle)
        from math import pi
        assert(maxAngle<=pi)
        assert(maxAngle>=minAngle)  # FIX: what really should be the maximum maxangle?
        assert(minAngle>=0)
        particles = make_particles_uniform_dist_range_angle(density,options.width,options.depth,
                                                          options.minLength,options.maxLength,
                                                          minAngle,maxAngle,
                                                          verbosity=options.verbosity)

    writearray('parts.dat',particles,True)

    if options.gpFile!=None:
        if options.verbosity>=TRACE:
            print 'Writing gnuplot data and command file to ', options.gpFile
        particles_gnuplot_dat(particles,options.width, 0, options.depth, options.gpFile,options.verbosity)
        #,options)

    if options.show:
        if options.verbosity>=TRACE:
            print 'starting show_particles to display figure'
        show_particles(density,options.width,options.depth, 1,3, particles,verbosity=options.verbosity)
    if options.plotFile!=None:
        if options.verbosity>=TRACE:
            print 'starting show_particles to save to: ',options.plotFile
        show_particles(density,options.width,options.depth, 1,3, particles,verbosity=options.verbosity,filename=options.plotFile)
